import argparse

from langchain.chat_models import ChatOpenAI

from DescKGC.procedures.extract_and_insert.base import (
    extract_entities_from_paper, get_paper_title_by_field)
from DescKGC.procedures.build_config import load_config
from DescKGC.tools.db_manager import DBManager
from DescKGC.tools.extractor.base import init_extract_chain


def add_arguments(parser):
    parser.add_argument("--temperature", type=float)


def main(args):
    config = load_config()
    if args.temperature is None:
        temperature = config["llm"]["temperature"]
    else:
        temperature = args.temperature
    llm = ChatOpenAI(temperature=temperature)
    extract_chain = init_extract_chain(llm)
    db_manager = DBManager(**config["neo4jdb"], **config["chromadb"])

    res = get_paper_title_by_field(db_manager, "doi")
    for title in res:
        try:
            extract_entities_from_paper(
                paper_id_type="title",
                paper_id_value=title,
                topic=config["topic"],
                shortenings=config["shortenings"],
                vs_key_info=config["extractor"]["entity"]["vs_key_info"],
                db_manager=db_manager,
                extract_chain=extract_chain,
            )
        except Exception as e:
            print(f"Error when extracting entities from {title}: {e}")
            continue


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    main(args)



from collections import OrderedDict

import oyaml as yaml

from .default_config import DescKGCConfig


class ConfigManager:
    def __init__(self, local_config_path=".kgc_config.yaml", default_config=DescKGCConfig(), verbose=False):
        self.local_config_path = local_config_path
        self.default_config = default_config

    def get_config(self):
        local_config = self._get_local_config(self.local_config_path)
        merged_config = DescKGCConfig(**local_config)
        # return ordered dict
        return merged_config.dict()

    def _convert_to_ordered_dict(self, config):
        if isinstance(config, dict):
            return OrderedDict([(key, self._convert_to_ordered_dict(value)) for key, value in config.items()])
        elif isinstance(config, list):
            return [self._convert_to_ordered_dict(value) for value in config]
        else:
            return config

    def _get_local_config(self, local_config_path):
        local_config = self._load_local_config(local_config_path)
        if local_config is None:
            print(
                "No local config found, use default config. "
                "Please run `desckgc manage_config --init` to initialize a local config."
            )
            exit(1)
        return local_config

    def init_local_config(self):
        local_config = self._load_local_config(self.local_config_path)
        if local_config is None:
            yaml.dump(self.default_config.dict(), open(self.local_config_path, "w"))
        else:
            print("Local config already exists, skip initialization.")

    def _load_local_config(self, local_config_path):
        try:
            with open(local_config_path, "r") as local_config_file:
                local_config = yaml.safe_load(local_config_file)
        except FileNotFoundError:
            local_config = None
        return local_config



from typing import List, Tuple, Dict, Any

from langchain.schema import BaseOutputParser


class EntityAlignOutputParser(BaseOutputParser):
    """Parse out comma separated lists."""

    @property
    def lc_serializable(self) -> bool:
        return True

    def get_format_instructions(self) -> str:
        return (
            "If you think the candidate entity 1 and 3 describe "
            "the same concept as the source entity, you should "
            "return `[1 <name of entity 1>,3 <name of entity 3>]`. "
            "If you think none of the candidate entities describe "
            "the same concept as the source entity, you should "
            "return `[]`"
        )

    def parse(self, text: str) -> List[Tuple]:
        """Parse the output of an LLM call."""
        entity_pair_list = []
        if text != "[]":
            entity_list = text[1:-1].strip().split(", ")
            entity_pair_list = [(int(entity.split(" ")[0]), " ".join(entity.split(" ")[1:])) for entity in entity_list]
        return entity_pair_list


class EntityMergeOutputParser(BaseOutputParser):
    """Parse out comma separated lists."""

    @property
    def lc_serializable(self) -> bool:
        return True

    def get_format_instructions(self) -> str:
        return "The result should be formatted as `<name>: <description> generalbility: <Ture|False>`."

    def parse(self, text: str) -> Dict[str, Any]:
        """Parse the output of an LLM call.

        :param text: the output of an LLM call
        :type text: str
        :return: the name
        :rtype: str
        """
        name, description_generalbility = text.split(": ", 1)
        description, generalbility_str = description_generalbility.split(" generalbility: ")
        if 'true' in generalbility_str.lower():
            generalbility = True
        elif 'false' in generalbility_str.lower():
            generalbility = False
        else:
            raise ValueError(f"generalbility_str should be either True or False, but got {generalbility_str}")
        return {"name": name, "description": description, "general": generalbility}



import argparse

from DescKGC.scripts import (add_from_arxiv, entity_alignment,
                             extract_entity_from_doc, manage_config, manage_db,
                             search_from_arxiv)


def main():
    parser = argparse.ArgumentParser(description="DescKGC CLI")
    subparsers = parser.add_subparsers(dest="command")

    module_manage_db = subparsers.add_parser("manage-db", help="Monitors and manages the database.")
    manage_db.add_arguments(module_manage_db)
    module_manage_db.set_defaults(func=manage_db.main)

    module_manage_config = subparsers.add_parser("manage-config", help="Monitors and manages the config file.")
    manage_config.add_arguments(module_manage_config)
    module_manage_config.set_defaults(func=manage_config.main)

    module_search_from_arxiv = subparsers.add_parser(
        "search-from-arxiv", help="Search papers from arXiv and dump them as json."
    )
    search_from_arxiv.add_arguments(module_search_from_arxiv)
    module_search_from_arxiv.set_defaults(func=search_from_arxiv.main)

    module_extract_entity_from_doc = subparsers.add_parser(
        "extract-entity-from-doc", help="extract entities from documents in the database."
    )
    extract_entity_from_doc.add_arguments(module_extract_entity_from_doc)
    module_extract_entity_from_doc.set_defaults(func=extract_entity_from_doc.main)

    module_add_from_arxiv = subparsers.add_parser("add-from-arxiv", help="Add papers from local json files to databases.")
    add_from_arxiv.add_arguments(module_add_from_arxiv)
    module_add_from_arxiv.set_defaults(func=add_from_arxiv.main)

    module_entity_alignment = subparsers.add_parser("entity-alignment", help="Align entities in the database.")
    entity_alignment.add_arguments(module_entity_alignment)
    module_entity_alignment.set_defaults(func=entity_alignment.main)

    args = parser.parse_args()
    if args.command is None:
        parser.print_help()
    else:
        args.func(args)


if __name__ == "__main__":
    main()



import uuid
from typing import Tuple

from DescKGC.tools.chroma.base import ChromaVectorStore
from DescKGC.tools.neo4j.base import Neo4jGraph
from .utils import validate_keys


class DBManager:
    def __init__(
        self,
        url,
        username,
        password,
        chroma_db_impl,
        persist_directory,
        collection_name,
        **kwargs,
    ):
        self.graph_db = self.init_graph_db(url, username, password, kwargs)
        self.vector_db = self.init_vector_db(chroma_db_impl, persist_directory, collection_name, kwargs)
        self.update_schema()

    def init_vector_db(self, chroma_db_impl: str, persist_directory: str, collection_name: str, kwargs) -> ChromaVectorStore:
        """Initialize the vector database.
        :param chroma_db_impl: the implementation of the vector database.
        :type chroma_db_impl: str
        :param persist_directory: the directory to store the vector database.
        :type persist_directory: str
        :param collection_name: the name of the collection.
        :type collection_name: str
        :param kwargs: the keyword arguments for the vector database.
        :type kwargs: dict
        :return: the vector database.
        :rtype: ChromaVectorStore
        """
        vector_db = ChromaVectorStore(
            chroma_db_impl=chroma_db_impl,
            persist_directory=persist_directory,
            collection_name=collection_name,
            **kwargs,
        )
        return vector_db

    def init_graph_db(self, url: str, username: str, password: str, kwargs) -> Neo4jGraph:
        """Initialize the graph database.
        :param url: the url of the graph database.
        :type url: str
        :param username: the username of the graph database.
        :type username: str
        :param password: the password of the graph database.
        :type password: str
        :param kwargs: the keyword arguments for the graph database.
        :type kwargs: dict
        :return: the graph database.
        :rtype: Neo4jGraph
        """
        graph_db = Neo4jGraph(url=url, username=username, password=password, **kwargs)
        return graph_db

    def close_driver(self):
        self.graph_db._driver.close()

    def add_from_arxiv(self, arxiv_papers, arxiv_prefix, vs_key_info):
        embedding_key = vs_key_info["embedding_key"]
        metadata_keys = vs_key_info["metadata_keys"]
        paper_keys = arxiv_papers[0].keys()
        expected_keys = [embedding_key, *metadata_keys]
        validate_keys(expected_keys, paper_keys)
        cypher_insturction_and_uuid_list = [self._arxiv_paper_to_cypher(paper, arxiv_prefix) for paper in arxiv_papers]

        for (cypher_insturction, uuid_), paper in zip(cypher_insturction_and_uuid_list, arxiv_papers):
            try:
                self.graph_db.query(cypher_insturction)
                metadata = {metadata_key: paper[metadata_key] for metadata_key in metadata_keys}
                metadata["embedding_source"] = embedding_key
                metadata["doc_source_type"] = "provided"
                metadata["type"] = "arxiv"
                self.vector_db.add(
                    documents=[paper[embedding_key]],
                    metadatas=[metadata],
                    ids=[uuid_],
                )
            except Exception as e:
                # TODO: make this Exception more specific
                print(f"arxiv paper insert error: {str(e)}")
        self._devide_author_from_arxiv_nodes()
        # self.update_schema()

    def _devide_author_from_arxiv_nodes(self) -> None:
        """extract author nodes from arxiv nodes in neo4j database"""
        self.graph_db._devide_author_from_arxiv_nodes()
        self.update_schema()

    def _arxiv_paper_to_cypher(self, arxiv_dict: dict, arxiv_prefix: str) -> Tuple[str, str]:
        """
        input: an arxiv res dict return from response_to_json
        output: a cypher CREATE instruction
        """
        uuid_ = f"{arxiv_prefix}-{uuid.uuid4()}"
        cypher_insturction = self.graph_db._get_arxiv_paper_insert_instruction(arxiv_dict, uuid_)
        return cypher_insturction, uuid_

    def show_schema(self) -> None:
        import json

        self.graph_db.refresh_schema()
        # print(self.graph_schema)
        print("Node properties are the following:")
        for node in self.node_properties:
            print(json.dumps(node, indent=4))

        print("\nRelationship properties are the following:")
        for relationship in self.rel_properties:
            print(json.dumps(relationship, indent=4))

        print("\nThe relationships are the following:")
        for rel in self.relationships:
            print(rel)

    def update_schema(self) -> None:
        self.graph_db.refresh_schema()
        self.graph_schema = self.graph_db.get_schema
        self.node_properties = self.graph_db.get_node_properties
        self.rel_properties = self.graph_db.get_rel_properties
        self.relationships = self.graph_db.get_relationships

    def delete_by_type(self, type: str) -> None:
        """
        delete all nodes of a type.
        TODO: implement delete by attribute
        """
        self.graph_db._delete_by_type(type)
        self.update_schema()

    def _get_entity_types_with_unique_prop(self, prop_name: str) -> list:
        """return a list of entity types that have unique prop_name"""
        SEARCH_INSTURCTION = f"""MATCH (n)
        WHERE n.{prop_name} is not NULL
        RETURN DISTINCT labels(n) AS labels"""
        entity_types = self.graph_db.query(SEARCH_INSTURCTION)
        entity_types = [entity_type["labels"][0] for entity_type in entity_types]
        return entity_types

    def get_node_labels(self):
        cypher_insturction = """CALL apoc.meta.data()
        YIELD label, other, elementType, type, property
        WHERE NOT type = "RELATIONSHIP" AND elementType = "node"
        WITH label AS nodeLabels, collect({property:property, type:type}) AS properties
        RETURN {labels: nodeLabels} AS output"""
        res = self.graph_db.query(cypher_insturction)
        return [label["output"]["labels"] for label in res]



from typing import Type

import chromadb
from chromadb.config import Settings
from chromadb.utils import embedding_functions


class ChromaVectorStore:
    def __init__(self, chroma_db_impl, persist_directory, collection_name, distance_function, **kwargs) -> None:
        self.client = chromadb.Client(
            Settings(
                chroma_db_impl=chroma_db_impl,
                persist_directory=persist_directory,
            )
        )
        self.collection_name = collection_name
        assert distance_function in [
            "cosine",
            "l2",
            "ip",
        ], """
        distance function should be one of cosine, l2, ip"""
        self.collection = self.client.get_or_create_collection(
            name=self.collection_name,
            embedding_function=embedding_functions.SentenceTransformerEmbeddingFunction(model_name="all-MiniLM-L6-v2"),
            metadata={"hnsw:space": distance_function},
        )
        # hnsw doc can be seen at
        # https://github.com/nmslib/hnswlib/tree/master#python-bindings.
        self.client.persist()

    def clear_collection(self):
        self.client.delete_collection(name=self.collection_name)
        self.client.persist()

    def add(self, documents, metadatas, ids):
        self.collection.add(documents=documents, metadatas=metadatas, ids=ids)
        self.client.persist()

    def get_by_uuid(self, uuid):
        return self.collection.get(
            ids=uuid,
        )

    def get_name_description_by_uuids(self, uuids):
        results = self.collection.get(
            ids=uuids
        )
        if results["metadatas"] is not None and results["documents"] is not None:
            names = [metadata["name"] for metadata in results["metadatas"]]
            descriptions = results["documents"]
            assert len(names) == len(descriptions)
        else:
            raise ValueError("uuid not in vector database.")
        pairs = list(zip(names, descriptions))
        return pairs

    def get_metadata_description_by_uuids(self, uuids):
        results = self.collection.get(
            ids=uuids
        )
        if results["metadatas"] is not None and results["documents"] is not None:
            descriptions = results["documents"]
            metadatas = results["metadatas"]
            assert len(metadatas) == len(descriptions)
        else:
            raise ValueError("uuid not in vector database.")
        pairs = list(zip(metadatas, descriptions))
        return pairs

    def get_from_specific_source(self, source):
        if source == "provided":
            embedding_source = "summary"
        elif source == "generated":
            embedding_source = "description"
        else:
            raise ValueError("embedding source should be either provided or generated")
        return self.collection.get(where={"embedding_source": embedding_source})

    def get_specific_type_nodes_uuid(self, node_type):
        nodes = self.collection.get(where={"type": node_type})
        return nodes["ids"]

    def query(self, query_texts, n_results, **kwargs):
        results = self.collection.query(query_texts=query_texts, n_results=n_results, **kwargs)
        # TODO: return ids and others necessary
        return results

    def query_from_specific_source(self, source, query_texts, n_results, **kwargs):
        if source == "provided":
            embedding_source = "summary"
        elif source == "generated":
            embedding_source = "description"
        else:
            raise ValueError("embedding source should be either provided or generated")
        results = self.collection.query(
            query_texts=query_texts, n_results=n_results, where={"embedding_source": embedding_source}, **kwargs
        )
        return results

    def query_from_specific_type(self, entity_type, query_texts, n_results, **kwargs):
        results = self.collection.query(
            query_texts=query_texts,
            n_results=n_results,
            where={"type": entity_type[0]},
            # TODO: make this `where` clause able to receive a list
            # by `or` operation.
            **kwargs
        )
        return results


ChromaVectorStoreType = Type[ChromaVectorStore]



from langchain.chains import LLMChain, SequentialChain
from langchain.prompts import PromptTemplate
from langchain.prompts.chat import (ChatPromptTemplate,
                                    HumanMessagePromptTemplate)

from .template import ENTITY_EXTRACT_TEMPLATE, RELATION_EXTRACT_TEMPLATE


def init_extract_chain(llm):
    entity_human_message_prompt = HumanMessagePromptTemplate(
        prompt=PromptTemplate(
            template=ENTITY_EXTRACT_TEMPLATE,
            input_variables=["topic", "summary"],
        )
    )
    entity_chat_prompt_template = ChatPromptTemplate.from_messages(
        [entity_human_message_prompt]
    )

    relation_human_message_prompt = HumanMessagePromptTemplate(
        prompt=PromptTemplate(
            template=RELATION_EXTRACT_TEMPLATE,
            input_variables=["summary", "entities"],
        )
    )
    relation_chat_prompt_template = ChatPromptTemplate.from_messages(
        [relation_human_message_prompt]
    )

    entity_extract_chain = LLMChain(
        llm=llm, prompt=entity_chat_prompt_template, output_key="entities"
    )

    relation_extract_chain = LLMChain(
        llm=llm, prompt=relation_chat_prompt_template, output_key="relations"
    )

    extract_chain = SequentialChain(
        chains=[entity_extract_chain, relation_extract_chain],
        input_variables=["topic", "summary"],
        output_variables=["entities", "relations"],
        verbose=True,
    )
    return extract_chain


if __name__ == "__main__":
    from langchain.chat_models import ChatOpenAI

    llm = ChatOpenAI(temperature=0.3)
    extract_chain = init_extract_chain(llm)
    summary = """
    Complex networks have attracted a great deal of research interest."""
    res = extract_chain({"summary": summary, "topic": "network science"})



from .utils import add_one_entity, add_one_relation, entity_relation_format


def extract_entities_from_paper(
    paper_id_type,
    paper_id_value,
    topic,
    shortenings,
    vs_key_info,
    db_manager,
    extract_chain,
):
    entity_list, relation_triple_list = entity_relation_format(paper_id_type, paper_id_value, topic, db_manager, extract_chain)
    if len(entity_list) == 0:
        return
    entity_id_dict = {}
    embedding_key = vs_key_info["embedding_key"]
    metadata_keys = vs_key_info["metadata_keys"]
    entity_keys = entity_list[0].keys()
    assert all(
        key in entity_keys for key in [embedding_key, *metadata_keys]
    ), """KeyError: some keys are not in the paper dict.
    Please check the vs_key_info in config.yaml"""
    for entity in entity_list:
        entity_shortning = shortenings[entity["type"]]
        entity_uuid = add_one_entity(db_manager, paper_id_type, paper_id_value, entity, entity_shortning)
        entity_id_dict[entity["name"]] = entity_uuid
        metadata = {metadata_key: entity[metadata_key] for metadata_key in metadata_keys}
        metadata["embedding_source"] = "description"
        metadata["doc_source_type"] = "generated"
        db_manager.vector_db.add(
            documents=[entity[embedding_key]],
            metadatas=[metadata],
            ids=[entity_uuid],
        )
    for relation_triple in relation_triple_list:
        add_one_relation(db_manager, relation_triple, entity_id_dict)


def extract_entities_from_papers(id_type, id_value_list, topic, db_manager, extract_chain, config):
    shortenings = config["shortenings"]
    vs_key_info = config["extractor"]["entity"]["vs_key_info"]
    for id_value in id_value_list:
        extract_entities_from_paper(
            id_type,
            id_value,
            topic,
            shortenings,
            vs_key_info,
            db_manager,
            extract_chain,
        )


def get_paper_title_by_field(db_manager, field_name):
    # return paper titles who have field_name
    query = """MATCH (p:Paper)
    WHERE p.{} IS NOT NULL
    RETURN p.title AS titles""".format(
        field_name
    )
    res = db_manager.graph_db.query(query)
    return [r["titles"] for r in res]



from typing import Type

from langchain.chains import LLMChain
from langchain.llms.base import BaseLLM
from langchain.prompts import PromptTemplate
from langchain.prompts.chat import ChatPromptTemplate, HumanMessagePromptTemplate, SystemMessagePromptTemplate

from .template import ALIGN_INPUT_TEMPLATE, ALIGN_TASK_FORMAT_TEMPLATE, MERGE_INPUT_TEMPLATE, MERGE_TASK_FORMAT_TEMPLATE


def init_align_chain(llm: Type[BaseLLM]) -> Type[LLMChain]:
    """Initialize the entity alignment chain.
    the entity alignment chain is used to align the entities with the same meaning.
    :param llm: the language model
    :type llm: Type[BaseLLM]
    :return: the entity alignment chain
    :rtype: Type[LLMChain]
    """
    system_message_prompt = SystemMessagePromptTemplate(
        prompt=PromptTemplate(
            template=ALIGN_TASK_FORMAT_TEMPLATE,
            input_variables=["topic"],
        )
    )
    human_message_prompt = HumanMessagePromptTemplate(
        prompt=PromptTemplate(
            template=ALIGN_INPUT_TEMPLATE,
            input_variables=["source_entity", "candidate_entities"],
        )
    )
    chat_prompt_template = ChatPromptTemplate.from_messages([system_message_prompt, human_message_prompt])
    align_chain = LLMChain(llm=llm, prompt=chat_prompt_template, output_key="entities")
    return align_chain


def init_entity_merge_chain(llm: Type[BaseLLM]) -> Type[LLMChain]:
    """Initialize the entity merge chain.
    the entity merge chain is used to merge the entities with the same meaning filtered by the entity alignment chain.

    :param llm: the language model
    :type llm: Type[BaseLLM]
    :return: the entity merge chain
    :rtype: Type[LLMChain]
    """
    system_message_prompt = SystemMessagePromptTemplate(
        prompt=PromptTemplate(
            template=MERGE_TASK_FORMAT_TEMPLATE,
            input_variables=["topic"],
        )
    )
    human_message_prompt = HumanMessagePromptTemplate(
        prompt=PromptTemplate(
            template=MERGE_INPUT_TEMPLATE,
            input_variables=["entities"],
        )
    )
    chat_prompt_template = ChatPromptTemplate.from_messages([system_message_prompt, human_message_prompt])
    merge_chain = LLMChain(llm=llm, prompt=chat_prompt_template, output_key="new_entity")
    return merge_chain


if __name__ == "__main__":
    from langchain.chat_models import ChatOpenAI

    from DescKGC.tools.align.utils import entities_nd_pair_warpper
    from DescKGC.tools.align.parser import EntityMergeOutputParser

    llm = ChatOpenAI(temperature=0.3)
    align_chain = init_align_chain(llm)
    merge_chain = init_entity_merge_chain(llm)
    merge_parser = EntityMergeOutputParser()
    src_entity = """Complex networks: have attracted a
    "great deal of research interest in the last two decades."""
    candidate_entities = [
        ("network science", "the study of complex networks"),
        (
            "network theory",
            "the study of graphs as a representation of either "
            "symmetric relations or asymmetric relations between"
            " discrete objects",
        ),
        (
            "Complex network",
            "a type of graph with non-trivial topological "
            "features—features that do not occur in simple "
            "networks such as lattices or random graphs but "
            "often occur in graphs modelling of real systems.",
        ),
    ]
    candidate_entities_fmt = entities_nd_pair_warpper(candidate_entities)
    res = align_chain(
        {
            "topic": "network science",
            "source_entity": src_entity,
            "candidate_entities": candidate_entities_fmt,
        }
    )
    print(res["entities"])
    new_entity = merge_chain({"topic": "network science", "entities": res["entities"]})
    print(merge_parser.parse(new_entity["new_entity"]))



import uuid


def generate_node_id(node_type):
    node_prefix = get_node_prefix(node_type)
    unique_id = str(uuid.uuid4())
    node_id = f"{node_prefix}{unique_id}"
    return node_id


def get_node_prefix(node_type):
    # TODO: import from configration file.
    prefixes = {
        "paper": "P",
        "concept": "C",
        "dataset": "D",
    }
    return prefixes.get(node_type.lower(), "")


if __name__ == "__main__":
    paper_node_id = generate_node_id("paper")
    print(paper_node_id)



from pydantic import BaseModel


class Neo4jDBSettings(BaseModel):
    url: str = "bolt://localhost:7687"
    username: str = "neo4j"
    password: str = "pleaseletmein"


class ChromaDBSettings(BaseModel):
    chroma_db_impl: str = "duckdb+parquet"
    persist_directory: str = "./data/chroma/"
    collection_name: str = "default_collection"
    distance_function: str = "l2"


class ArxivExtractorSettings(BaseModel):
    data_path: str = "./data/arxiv/"
    vs_key_info: dict = {"embedding_key": "summary", "metadata_keys": ["title"]}


class EntityExtractorSettings(BaseModel):
    vs_key_info: dict = {"embedding_key": "description", "metadata_keys": ["name", "general", "type"]}


class LLMSettings(BaseModel):
    model_name: str = "gpt-3.5-turbo"
    temperature: float = 0.3


class ShorteningsSettings(BaseModel):
    Paper: str = "p"
    Concept: str = "c"
    Model: str = "m"
    Dataset: str = "d"


class EntityAlignmentSettings(BaseModel):
    entity_types: list = ["Concept", "Model", "Dataset"]
    threshold: float = 0.08


class DescKGCConfig(BaseModel):
    topic: str = "network science"
    neo4jdb: Neo4jDBSettings = Neo4jDBSettings()
    chromadb: ChromaDBSettings = ChromaDBSettings()
    extractor: dict = {"arxiv": ArxivExtractorSettings(), "entity": EntityExtractorSettings()}
    llm: LLMSettings = LLMSettings()
    shortenings: ShorteningsSettings = ShorteningsSettings()
    entity_alignment: EntityAlignmentSettings = EntityAlignmentSettings()



from typing import List, Tuple, Type, Dict

from langchain.chains import LLMChain
from langchain.schema import BaseOutputParser

from DescKGC.tools.align.utils import entities_nd_pair_warpper, entities_ndg_pair_warpper
from DescKGC.tools.db_manager.type import DBManagerType


def get_entity_type_uuids(db_manager: DBManagerType, entity_types: list) -> Dict:
    """Return a dictionary of entity types and their uuids.

    :param db_manager: DBManager object
    :type db_manager: DBManagerType
    :param entity_types: list of entity types
    :type entity_types: list
    :return: dictionary of entity types and their uuids
    :rtype: Dict
    """
    assert isinstance(entity_types, list), "entity_types should be a list"
    entity_type_uuids = {node_type: [] for node_type in entity_types}
    for node_type in entity_types:
        uuids = db_manager.vector_db.get_specific_type_nodes_uuid(node_type)
        entity_type_uuids[node_type] = uuids
    return entity_type_uuids


def query_from_specific_type_uuids(
    db_manager: DBManagerType,
    target_entity_types: list,
    query_entity_uuid: str,
) -> list:
    """Return a list of similar entities from the target entity types.

    given a query entity uuid.
    :param db_manager: DBManager object
    :type db_manager: DBManagerType
    :param target_entity_types: list of target entity types
    :type target_entity_types: list
    :param query_entity_uuid: uuid of the query entity
    :type query_entity_uuid: str
    :return: list of similar entities
    :rtype: list
    """
    query = db_manager.vector_db.get_by_uuid(uuid=[query_entity_uuid])["documents"][0]

    results = db_manager.vector_db.query_from_specific_type(target_entity_types, query, n_results=10)
    return results


def select_candidate_entities_uuids(threshold: float, similar_entities: dict, src_entity_uuid: str) -> list:
    """Return a list of the uuids of similar entities within the threshold.

    distance from the source entity.
    :param threshold: threshold distance
    :type threshold: float
    :param similar_entities: similar entities
    :type similar_entities: dict
    :param src_entity_uuid: uuid of the source entity
    :type src_entity_uuid: str
    :return: list of uuids of similar entities within the threshold distance
    :rtype: list
    """

    distances = similar_entities["distances"][0]
    entity_uuids = similar_entities["ids"][0]

    # Remove source entity from options
    if entity_uuids[0] == src_entity_uuid:
        entity_uuids.pop(0)
        distances.pop(0)

    candidate_uuids = []
    for uuid, dist in zip(entity_uuids, distances):
        if dist <= threshold:
            candidate_uuids.append(uuid)
    return candidate_uuids


def align_source_and_candidates_with_chain(
    topic: str,
    db_manager: DBManagerType,
    align_chain: Type[LLMChain],
    align_parser: Type[BaseOutputParser],
    uuid: str,
    candidate_uuids: List[str],
) -> List[Tuple[int, str]]:
    """use the align_chain to align the source entity and the candidate entities.

    :param topic: The topic of the conversation
    :type topic: str
    :param db_manager: The database manager
    :type db_manager: DBManagerType
    :param align_chain: The align chain
    :type align_chain: Type[LLMChain]
    :param align_parser: The align parser
    :type align_parser: Type[BaseOutputParser]
    :param uuid: The uuid of the source entity
    :type uuid: str
    :param candidate_uuids: The uuids of the candidate entities
    :type candidate_uuids: List[str]
    :return: A list of tuples, each tuple contains the order of the entity and the entity name.
        For example, [(0, "entity1"), (1, "entity2")].
    :rtype: List[Tuple[int, str]]
    """

    source_entity_pair = db_manager.vector_db.get_name_description_by_uuids(uuid)
    source_entity_pair_fmt = entities_nd_pair_warpper(source_entity_pair, is_candiate=False)
    target_entity_pairs = db_manager.vector_db.get_name_description_by_uuids(candidate_uuids)
    target_entity_pairs_fmt = entities_nd_pair_warpper(target_entity_pairs, is_candiate=True)
    res = align_chain(
        {
            "topic": topic,
            "source_entity": source_entity_pair_fmt,
            "candidate_entities": target_entity_pairs_fmt,
        }
    )
    entities = align_parser.parse(res["entities"])
    return entities


def merge_entities_with_chain(
    db_manager: DBManagerType,
    topic: str,
    merge_chain: Type[LLMChain],
    merge_parser: Type[BaseOutputParser],
    uuid: str,
    selected_candidate_uuids: List[str],
) -> Tuple[str, str]:
    """use the merge_chain to merge the source entity and the candidate entities.

    :param db_manager: The database manager
    :type db_manager: DBManagerType
    :param topic: The topic of the knowledge graph
    :type topic: str
    :param merge_chain: The merge chain
    :type merge_chain: Type[LLMChain]
    :param merge_parser: The merge parser
    :type merge_parser: Type[BaseOutputParser]
    :param uuid: The uuid of the source entity
    :type uuid: str
    :param selected_candidate_uuids: The uuids of the candidate entities
    :type selected_candidate_uuids: List[str]
    :return: The name and description of the new entity
    :rtype: Tuple[str, str]
    """
    # name_descriptions_to_merge = db_manager.vector_db.get_name_description_by_uuids([uuid] + selected_candidate_uuids)
    metadata_descriptions_to_merge = db_manager.vector_db.get_metadata_description_by_uuids([uuid] + selected_candidate_uuids)
    ndg_input = [(entity[0]['name'], entity[1], entity[0]['general']) for entity in metadata_descriptions_to_merge]
    ndg_input_fmt = entities_ndg_pair_warpper(ndg_input)
    # target_entity_pairs_fmt = entities_nd_pair_warpper(ndg_input_fmt)
    res = merge_chain(
        {
            "topic": topic,
            "entities": ndg_input_fmt,
        }
    )
    new_entity = merge_parser.parse(res["new_entity"])
    return new_entity



from typing import Any, Dict, List

from neo4j import Query

from .cypher_template import (ARXIV_PAPER_INSERT_INSTRUCTION,
                              DELETE_NODES_INSTRUCTION,
                              DETACH_AUTHOR_FROM_PAPER_INSTRUCTION)
from .utils import join_if_list

node_properties_query = Query("""
CALL apoc.meta.data()
YIELD label, other, elementType, type, property
WHERE NOT type = "RELATIONSHIP" AND elementType = "node"
WITH label AS nodeLabels, collect({property:property, type:type}) AS properties
RETURN {labels: nodeLabels, properties: properties} AS output
""")

rel_properties_query = Query("""
CALL apoc.meta.data()
YIELD label, other, elementType, type, property
WHERE NOT type = "RELATIONSHIP" AND elementType = "relationship"
WITH label AS nodeLabels, collect({property:property, type:type}) AS properties
RETURN {type: nodeLabels, properties: properties} AS output
""")

relationships_query = Query("""
CALL apoc.meta.data()
YIELD label, other, elementType, type, property
WHERE type = "RELATIONSHIP" AND elementType = "node"
RETURN "(:" + label + ")-[:" + property + "]->(:" + toString(other[0]) + ")"
AS output
""")


class Neo4jGraph:
    """Neo4j wrapper for graph operations."""

    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        database: str = "neo4j",
        **kwargs,
    ) -> None:
        """Create a new Neo4j graph wrapper instance."""
        try:
            import neo4j
        except ImportError:
            raise ValueError(
                "Could not import neo4j python package. "
                "Please install it with `pip install neo4j`."
            )

        self._driver = neo4j.GraphDatabase.driver(
            url, auth=(username, password)
        )
        self._database = database
        self.schema = ""
        self.node_properties = {}
        self.rel_properties = {}
        self.relationships = []
        try:
            self.refresh_schema()
        except neo4j.exceptions.ClientError:
            raise ValueError(
                "Could not use APOC procedures. "
                "Please install the APOC plugin in Neo4j."
            )

    @property
    def get_schema(self) -> str:
        """Returns the schema of the Neo4j database"""
        return self.schema

    @property
    def get_node_properties(self) -> List[Dict[str, Any]]:
        """Returns the node properties of the Neo4j database"""
        return self.node_properties

    @property
    def get_rel_properties(self) -> List[Dict[str, Any]]:
        """Returns the relationship properties of the Neo4j database"""
        return self.rel_properties

    @property
    def get_relationships(self) -> List[str]:
        """Returns the relationships of the Neo4j database"""
        return self.relationships

    def query(self, query: Query, params: dict = {}) -> List[Dict[str, Any]]:
        """Query Neo4j database."""
        from neo4j.exceptions import CypherSyntaxError

        with self._driver.session(database=self._database) as session:
            try:
                data = session.run(query, params)
                return [r.data() for r in data]
            except CypherSyntaxError as e:
                raise ValueError(
                    "Generated Cypher Statement is not valid\n" f"{e}"
                )

    def refresh_schema(self) -> None:
        """
        Refreshes the Neo4j graph schema information.
        """
        node_properties = self.query(node_properties_query)
        relationships_properties = self.query(rel_properties_query)
        relationships = self.query(relationships_query)

        self.schema = f"""
        Node properties are the following:
        {[el['output'] for el in node_properties]}
        Relationship properties are the following:
        {[el['output'] for el in relationships_properties]}
        The relationships are the following:
        {[el['output'] for el in relationships]}
        """
        self.node_properties = [el["output"] for el in node_properties]
        self.rel_properties = [el["output"] for el in relationships_properties]
        self.relationships = [el["output"] for el in relationships]

    def _devide_author_from_arxiv_nodes(self) -> None:
        cypher_insturction = DETACH_AUTHOR_FROM_PAPER_INSTRUCTION
        self.query(cypher_insturction)

    def _delete_by_type(self, type: str) -> None:
        cypher_insturction = DELETE_NODES_INSTRUCTION.format(type=type)
        self.query(cypher_insturction)

    def _get_arxiv_paper_insert_instruction(self, arxiv_dict, uuid_):
        return ARXIV_PAPER_INSERT_INSTRUCTION.format(
            title=join_if_list(arxiv_dict["title"]),
            authors=join_if_list(arxiv_dict["authors"]),
            published=join_if_list(arxiv_dict["published"]),
            updated_date=join_if_list(arxiv_dict["updated_date"]),
            summary=join_if_list(arxiv_dict["summary"].replace("\n", " ")).replace('"', '`'),
            doi=join_if_list(arxiv_dict["doi"]),
            primary_category=join_if_list(arxiv_dict["primary_category"]),
            categories=join_if_list(arxiv_dict["categories"]),
            pdf_url=join_if_list(arxiv_dict["pdf_url"]),
            uuid=uuid_,
        )

    def _show_current_index(self) -> list:
        """return a list of index names"""
        cypher_insturction = """SHOW INDEXES"""
        index_list = self.query(cypher_insturction)
        index_list = [index["name"] for index in index_list]
        return index_list

    def create_or_update_index_on_unique_property(self, property_name: str) -> None:
        """update index for a given attribute"""
        if f"{property_name}_index" in self._show_current_index():
            cypher_drop_insturction = f"""DROP INDEX {property_name}_index"""
            self.query(cypher_drop_insturction)
        entity_types = self._get_entity_types_with_unique_prop(property_name)
        str_entity_types = " | ".join(entity_types)
        cypher_insturction = f"""CREATE FULLTEXT INDEX {property_name}_index
        FOR (n:{str_entity_types})
        ON EACH [n.{property_name}]"""
        self.query(cypher_insturction)

    def search_by_index(self, property_name: str, search_string: str) -> list:
        """search by index"""
        cypher_insturction = f"""CALL db.index.fulltext.queryNodes("{property_name}_index", "{search_string}")
        YIELD node, score
        RETURN node.title, score"""
        res = self.query(cypher_insturction)
        # convert to list of pairs
        res = [(pair["node.title"], pair["score"]) for pair in res]
        print(res)
        return res



ALIGN_TASK_FORMAT_TEMPLATE = """You are a domain expert in the field of {topic}. During a knowledge graph construction process, you are asked to discriminate if the candidate entity(s) describe the same concept as the source entity so that we should merge them into one entity. Each entity contains the name and description in the format of `<name>: <description>`.
Besides each candidate entity is attached with a number to make it easy to refer to. Please read the source entity and candidate entities carefully and select the correct candidate entity(s) that describe the same concept as the source entity.
The result should be formatted as a list of numbers and names separated by comma. For example, if you think the candidate entity 1 and 3 describe the same concept as the source entity, you should return `[1 <name of entity 1>,3 <name of entity 3>]`. If you think none of the candidate entities describe the same concept as the source entity, you should return `[]`.

For example, given the source entity:
"Complex netwrok: Networks with non-trivial topological features and patterns of connection between nodes."
and the candidate entities:
[
    "1. complex network: non-trivial topological features and patterns of connection between nodes.",
    "2. small world network: a network in which most nodes are not neighbors of one another, but most nodes can be reached from every other by a small number of hops or steps.",
    "3. network science: an academic field which studies complex networks such as telecommunication networks, computer networks, biological networks, cognitive and semantic networks, and social networks."
]
You should return `[1 complex network]` because the candidate entity 1 describes the same concept as the source entity and the candidate entity 2 and 3 do not describe the same concept as the source entity."""

ALIGN_INPUT_TEMPLATE = """the source entity:
{source_entity}
the candidate entities:
{candidate_entities}
your answer:"""

MERGE_TASK_FORMAT_TEMPLATE = """You are a domain expert in the field of {topic}. During a knowledge graph construction process, you are asked to merge the entities with the same meaning to reduce the abandoned entities.
Each entity contains the name and description in the format of `<name>: <description> generalbility: <True|False>`. The `generalbility` indicates whether the entity is general or not from previous knowledge graph construction process. You should merge the entities with the same meaning into one entity and rethink the generalbility of the new entity. If the new entity is not general, you should set the `generalbility` to `False` and it will be deleted from the knowledge graph because I only want to keep the domain-specific and well-defined entities.
Please read the entities carefully and write a new entity that describes the same concept as the entities you read.
The result should be formatted as `<name>: <description> generalbility: <Ture|False>`."""

MERGE_INPUT_TEMPLATE = """the entities:
{entities}
your answer:"""


from typing import List, Union


def response_to_json(search_res):
    res = []
    for result in search_res.results():
        res.append({
            "title": result.title,
            "authors": [author.name for author in result.authors],
            "published": str(result.published),
            "updated_date": str(result.updated),
            "summary": result.summary,
            "doi": result.doi,
            "primary_category": result.primary_category,
            "categories": result.categories,
            "pdf_url": result.pdf_url,
        })
    return res


def join_if_list(input: Union[List[str], str, int]) -> str:
    if isinstance(input, list):
        return "||".join(input)
    else:
        return str(input)



import argparse

from DescKGC.procedures.build_config import load_config
from DescKGC.tools.db_manager import DBManager


def add_arguments(parser):
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--delete-by-type",
        action="store_true",
        help="Delete nodes by type. You will be asked to enter the type.",
    )
    group.add_argument(
        "--delete-all",
        action="store_true",
        help="Delete all nodes. You will be asked to confirm.",
    )
    group.add_argument(
        "--show-schema",
        action="store_true",
        help="Show the schema of the database.",
    )


def main(args):
    config = load_config()

    db_manager = DBManager(**config["neo4jdb"], **config["chromadb"])
    node_labels = db_manager.get_node_labels()

    if args.delete_by_type:
        delete_by_type(db_manager, node_labels)

    if args.delete_all:
        delete_all(db_manager, node_labels)

    if args.show_schema:
        show_schema(db_manager)


def show_schema(db_manager):
    db_manager.show_schema()


def delete_all(db_manager, node_labels):
    if len(node_labels) == 0:
        print("No data in the database.")
    else:
        confirmation = input("Are you sure to delete all data? (y/n): ")
        if confirmation == "y":
            for node_label in node_labels:
                db_manager.delete_by_type(node_label)
            db_manager.vector_db.client.reset()
        else:
            print("Delete aborted.")


def delete_by_type(db_manager, node_labels):
    if len(node_labels) == 0:
        print("No data in the database.")
    else:
        print("The available types are:", end=" ")
        print(f'{", ".join(node_labels)}.')
        type_name = input("Please input the type name: \n")
        db_manager.delete_by_type(type_name)
        print(f"Delete all nodes with type {type_name}.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    main(args)



import argparse
import json
from pathlib import Path

from DescKGC.procedures.build_config import load_config
from DescKGC.tools.db_manager import DBManager


def add_arguments(parser):
    parser.add_argument("--data-path", type=str)
    parser.add_argument("--file-name", type=str, required=True)


def main(args):
    config = load_config()
    if args.data_path is None:
        data_path = config["extractor"]["arxiv"]["data_path"]
    else:
        data_path = args.data_path
    file_name = args.file_name

    # Load the json file
    data_path = Path(data_path)
    file_path = data_path / f"{file_name}.json"

    # Initiate a neo4j manager
    db_manager = DBManager(**config["neo4jdb"], **config["chromadb"])

    print(f"adding data from {file_path}")
    with open(file_path, "r") as f:
        arxiv_papers = json.load(f)
    db_manager.add_from_arxiv(
        arxiv_papers,
        arxiv_prefix=config["shortenings"]["Paper"],
        vs_key_info=config["extractor"]["arxiv"]["vs_key_info"],
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    main(args)



import json
import re
import time
import uuid


def convert_to_triples(input_string):
    # Use regular expressions to find all substrings that look like '( ... )'
    triple_strings = re.findall(r"\((.*?)\)", input_string)
    triples = []
    for string in triple_strings:
        # Split by ', ' and append as a tuple
        triple = string.split(", ")
        triples.append(tuple(triple))
    return triples


def entity_relation_format(
    id_type, id_value, topic, db_manager, extract_chain
):
    "Return a list of entities and a list of relations"
    summary = db_manager.graph_db.query(
        f"""MATCH (n:Paper) WHERE n.{id_type}=$id_value
        RETURN n.summary as summary""",
        params={"id_value": id_value},
    )
    res = extract_chain({"topic": topic, "summary": summary})
    try:
        entity_list = json.loads(res["entities"])
    except json.decoder.JSONDecodeError as e:
        print(f"Failed to decode JSON string: {str(e)}")
        return [], []
    try:
        relation_list = convert_to_triples(res["relations"])
    except Exception as e:
        print(f"Failed to decode triples from relations: {str(e)}")
        return [], []
    return entity_list, relation_list


def add_one_entity(db_manager, id_type, id_value, entity, shortning):
    current_time = time.localtime()
    time_str = time.strftime("%Y-%m-%d %H:%M:%S", current_time)
    uuid_ = f"{shortning}-{uuid.uuid4()}"
    _ = db_manager.graph_db.query(
        f"""
            MATCH (p:Paper) WHERE p.{id_type}=$id_value
            WITH p
            MERGE (c:{entity['type']} {{name:$name, description:$description,
            general:$general, uuid:$uuid}})
            MERGE (c)-[r:EXTRACTED_FROM {{timestep:$timestep}}]->(p)
            RETURN elementid(c) AS elementid
            """,
        params={
            "id_value": id_value,
            "name": entity["name"],
            "description": entity["description"],
            "general": entity["general"],
            "timestep": time_str,
            "uuid": uuid_,
        },
    )
    return uuid_


def add_one_relation(db_manager, relation_triple, entity_uuid_dict):
    current_time = time.localtime()
    time_str = time.strftime("%Y-%m-%d %H:%M:%S", current_time)
    try:
        src_uuid, dst_uuid = (
            entity_uuid_dict[relation_triple[0]],
            entity_uuid_dict[relation_triple[2]],
        )
    except Exception as e:
        print(f"Error: Inconsistent Entity Extraction - {str(e)}")
        print(
            """Details: The entities present in the relation triples were not
            previously added as nodes during the entity extraction phase."""
        )
        print(
            """Suggestion: Ensure that the Language Model maintains consistency
            between the entity extraction and relation extraction steps."""
        )
        return
    relation_name = relation_triple[1].upper().replace(" ", "_")
    db_manager.graph_db.query(
        f"""
            MATCH (src) WHERE src.uuid=$src_uuid
            MATCH (dst) WHERE dst.uuid=$dst_uuid
            MERGE (src)-[rel: {relation_name} {{timestep: $timestep}}]->(dst)
            """,
        params={
            "src_uuid": src_uuid,
            "dst_uuid": dst_uuid,
            "timestep": time_str,
        },
    )



import argparse

import arxiv

from DescKGC.procedures.build_config import load_config
from DescKGC.tools.arxiv.base import dump_to_json, response_to_dataclass


def add_arguments(parser):
    parser.add_argument("--query", type=str, required=True)
    parser.add_argument("--max-results", type=int, default=4)
    parser.add_argument("--data-path", type=str)


def main(args):
    config = load_config()
    # convert query as a argument
    query = args.query
    max_results = args.max_results
    if args.data_path is None:
        data_path = config["extractor"]["arxiv"]["data_path"]
    else:
        data_path = args.data_path
    search = arxiv.Search(
        query=query,
        max_results=max_results,
        sort_by=arxiv.SortCriterion.Relevance,
        sort_order=arxiv.SortOrder.Descending,
    )
    res = response_to_dataclass(search)
    dumped_file_name = query.replace(" ", "_").replace("-", "_")
    dump_to_json(res, data_path, dumped_file_name)
    print(f"dumped to {data_path} as {dumped_file_name}.json")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    main(args)



ENTITY_EXTRACT_TEMPLATE = """You are provided with a research paper summary. Your task is to extract key entities from the summary that are related to a given topic. These entities should belong to one of the following categories:

- Concept: A concept or idea presented in the paper, e.g., "deep learning", "convolutional neural network", "attention mechanism".
- Model: A specific model discussed in the paper, e.g., "BERT", "ResNet", "Transformer".
- Dataset: A dataset used or referenced in the paper, e.g., "MNIST", "CIFAR-10", "ImageNet".

Extract all relevant entities from the summary and return them as a list. Each entity should be represented in JSON format with these keys:

- type: The category of the entity (i.e., "Concept", "Model", "Dataset").
- name: The name of the entity.
- description: A brief description of the entity, sourced from the summary.
- general: A boolean indicating if the entity is a general concept within the given topic or unique to the specific paper. For example, "language model" is a general concept within the topic "Natural Language Processing", while "a dataset of 1000 images of cats" is specific to a particular paper.


Here's an example. Given the summary "BERT is a language model proposed by Google in 2018. It is based on the Transformer architecture." and the topic "Natural Language Processing", the output should be:

[
  {{"type": "Model", "name": "BERT", "description": "BERT is a language model proposed by Google in 2018", "general": true}},
  {{"type": "Concept", "name": "Transformer", "description": "Transformer is a neural network architecture", "general": true}},
  {{"type": "Concept", "name": "language model", "description": "A language model predicts the next word given the previous words", "general": true}},
  {{"type": "Concept", "name": "neural network", "description": "A neural network is a model that learns from data", "general": true}}
]

Note: The order of entities does not matter. Do not use external resources for descriptions; all descriptions should be derived from the given summary.

Now, proceed with the following inputs:

Topic:
{topic}

Summary:
{summary}

Output Entities:

"""

RELATION_EXTRACT_TEMPLATE = """Given a research paper summary and a list of entities within the summary, your task is to extract and identify the relations between these entities.

The relations should fall into one of these categories:

- alias: Indicates that two entities are essentially the same, e.g., "BERT" and "Bidirectional Encoder Representations from Transformers".
- is superior to: Indicates a hierarchical relationship where one concept is more abstract or general than the other. For example, "language model" is superior to "BERT".
- based on: Indicates that one entity is based on another. For example, "BERT" is based on "Transformer".

The output should consist of triples: (entity1, relation, entity2), where entity1 and entity2 come from the provided entities, and relation represents a DIRECTED relationship between the two.

For instance, if the summary reads:
"BERT, also known as Bidirectional Encoder Representations from Transformers, is a language model proposed by Google in 2018. It is based on the Transformer architecture.",
and the entities are:
[
  {{"type": "Model", "name": "BERT", "description": "BERT is a language model proposed by Google in 2018", "general": true}},
  {{"type": "Model", "name": "Transformer", "description": "Transformer is a neural network architecture proposed by Google in 2017", "general": true}},
  {{"type": "Concept", "name": "language model", "description": "A language model is a model that can predict the next word given the previous words", "general": true}},
]

then the output should be:
(BERT, alias, Bidirectional Encoder Representations from Transformers)
(BERT, based on, Transformer)
(language model, is superior to, BERT)

Note: The description of relations should come directly from the summary and be as concise as possible. Do not use external resources. The DIRECTION of the relations is critical.

Now let's get started!

The input summary is

{summary}

The input entities are

{entities}

The output relations are
"""



import dataclasses
import json

from .arxiv_data import AriXivData, AriXivDataType, List


def response_to_dataclass(search_res) -> List[AriXivDataType]:
    res = []
    for result in search_res.results():
        search_result = AriXivData(
            title=result.title,
            authors=[author.name for author in result.authors],
            published=str(result.published),
            updated_date=str(result.updated),
            summary=result.summary,
            doi=result.doi,
            primary_category=result.primary_category,
            categories=result.categories,
            pdf_url=result.pdf_url,
        )
        res.append(search_result)
    return res


def dump_to_json(res: List[AriXivDataType], dir_path: str, file_name: str):
    file_path = dir_path + file_name + ".json"
    with open(file_path, "w") as f:
        json.dump([dataclasses.asdict(item) for item in res], f)



from DescKGC.tools.config_manager import ConfigManager


def add_arguments(parser):
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--init",
        action="store_true"
    )


def main(args):
    config_manager = ConfigManager()
    if args.init:
        config_manager.init_local_config()



ARXIV_PAPER_INSERT_INSTRUCTION = """MERGE (:Paper {{title: "{title}", authors: "{authors}", published: "{published}", updated_date: "{updated_date}", summary: "{summary}", doi: "{doi}", primary_category: "{primary_category}", categories: "{categories}", pdf_url: "{pdf_url}", uuid: "{uuid}"}})"""

DELETE_NODES_INSTRUCTION = """MATCH (n:{type})
DETACH DELETE n"""

DETACH_AUTHOR_FROM_PAPER_INSTRUCTION = """
match (p:Paper)
WITH p, split(p.authors, '||') AS authors
UNWIND authors as author
MERGE (a:Author {name: author})
MERGE (a)-[r:WROTE {timestep:p.published}]->(p)
"""



from typing import List, Tuple


def entities_nd_pair_warpper(candidate_entities: List[tuple], is_candiate=True) -> str:
    """Wrap the candidate entities into a string.
    The input is a list of tuples, each tuple contains the name and the description of the entity.
    :param candidate_entities: The candidate entities
    :type candidate_entities: list[tuple]
    :param is_candiate: Whether the candidate entities are the source entity, defaults to True
    :type is_candiate: bool, optional
    :return: The wrapped string
    :rtype: str
    """
    if not is_candiate:
        assert len(candidate_entities) == 1
        prefix = ""
        suffix = ""
    else:
        if len(candidate_entities) == 0:
            return "[]"
        prefix = "[\n"
        suffix = "\n]"
    content = ""
    for name, description in candidate_entities:
        content += f"""\t"{name}: {description}",\n"""
    return prefix + content[:-2] + suffix


def entities_ndg_pair_warpper(candidate_entities: List[Tuple[str, str, bool]], is_candidate=True) -> str:
    """Wrap the candidate entities into a string.
    The input is a list of tuples, each tuple contains the name, the description and generalbility of the entity.
    :param candidate_entities: The candidate entities
    :type candidate_entities: list[tuple]
    :param is_candiate: Whether the candidate entities are the source entity, defaults to True
    :type is_candiate: bool, optional
    """
    if not is_candidate:
        raise NotImplementedError
    else:
        if len(candidate_entities) == 0:
            return "[]"
        prefix = "[\n"
        suffix = "\n]"
    content = ""
    for name, description, generalbility in candidate_entities:
        content += f"""\t"{name}": "{description}" """
        content += f""" generalbility: {generalbility},\n"""
    return prefix + content[:-2] + suffix


if __name__ == "__main__":
    formated_nd = entities_nd_pair_warpper([("Q1", "A"), ("Q2", "B")])
    format_ndg = entities_ndg_pair_warpper([("Q1", "A", True), ("Q2", "B", False)])
    print(formated_nd)
    print(format_ndg)



from DescKGC.tools.config_manager import ConfigManager


def load_config():
    """Load configuration file from the current directory.
    Currently, the configuration file is named as `.kgc_config.yaml`.
    """
    config_manager = ConfigManager()
    config = config_manager.get_config()
    return config



from dataclasses import dataclass
from typing import List, Type


@dataclass
class AriXivData:
    title: str
    authors: List[str]
    published: str
    updated_date: str
    summary: str
    doi: str
    primary_category: str
    categories: List[str]
    pdf_url: str


AriXivDataType = Type[AriXivData]



import time
import uuid


def add_one_merged_entity(db_manager, id_type, id_values, entity, shortning):
    current_time = time.localtime()
    time_str = time.strftime("%Y-%m-%d %H:%M:%S", current_time)
    uuid_ = f"{shortning}-{uuid.uuid4()}"
    _ = db_manager.graph_db.query(
        f"""
            MATCH (e) WHERE e.{id_type} IN $id_values
            WITH e
            MERGE (c:{entity['type']} {{name:$name, description:$description,
            general:$general, uuid:$uuid}})
            MERGE (c)-[r:MERGED_FROM {{timestep:$timestep}}]->(e)
            RETURN elementid(c) AS elementid
            """,
        params={
            "id_values": id_values,
            "name": entity["name"],
            "description": entity["description"],
            "general": entity["general"],
            "timestep": time_str,
            "uuid": uuid_,
        },
    )
    return uuid_



import argparse

from langchain.chat_models import ChatOpenAI

from DescKGC.procedures.align_across_subgraphs.base import (
    align_source_and_candidates_with_chain,
    get_entity_type_uuids,
    query_from_specific_type_uuids,
    select_candidate_entities_uuids,
    merge_entities_with_chain,
)
from DescKGC.procedures.build_config import load_config
from DescKGC.tools.align.base import init_align_chain, init_entity_merge_chain
from DescKGC.tools.align.parser import EntityAlignOutputParser, EntityMergeOutputParser
from DescKGC.tools.db_manager import DBManager
from DescKGC.procedures.align_across_subgraphs.utils import add_one_merged_entity


def add_arguments(parser):
    parser.add_argument(
        "--entity_types",
        type=list,
    )


def main(args):
    config = load_config()
    if args.entity_types is None:
        entity_types = config["entity_alignment"]["entity_types"]
    else:
        entity_types = args.entity_types
    shortenings = config["shortenings"]
    metadata_keys = config["extractor"]["entity"]["vs_key_info"]["metadata_keys"]
    embedding_key = config["extractor"]["entity"]["vs_key_info"]["embedding_key"]

    db_manager = DBManager(**config["neo4jdb"], **config["chromadb"])
    llm = ChatOpenAI(temperature=config["llm"]["temperature"])
    topic = config["topic"]
    align_chain = init_align_chain(llm=llm)
    align_parser = EntityAlignOutputParser()
    merge_chain = init_entity_merge_chain(llm=llm)
    merge_parser = EntityMergeOutputParser()
    entity_type_uuids_dict = get_entity_type_uuids(db_manager, entity_types)

    for entity_type, uuids in entity_type_uuids_dict.items():
        for uuid in uuids:
            try:
                similar_entities = query_from_specific_type_uuids(db_manager, [entity_type], uuid)
                candidate_uuids = select_candidate_entities_uuids(
                    threshold=config["entity_alignment"]["threshold"],
                    similar_entities=similar_entities,
                    src_entity_uuid=uuid,
                )
                if len(candidate_uuids) > 0:
                    selected_entities = align_source_and_candidates_with_chain(
                        topic, db_manager, align_chain, align_parser, uuid, candidate_uuids
                    )
                    selected_entities_ids = [entity[0] - 1 for entity in selected_entities]
                    selected_candidate_uuids = [candidate_uuids[i] for i in selected_entities_ids]
                    new_entity = merge_entities_with_chain(
                        db_manager, topic, merge_chain, merge_parser, uuid, selected_candidate_uuids
                    )
                    new_entity["type"] = entity_type
                    entity_shortning = shortenings[entity_type]
                    uuids_to_link = [uuid] + selected_candidate_uuids
                    uuid_ = add_one_merged_entity(
                        db_manager=db_manager,
                        id_type="uuid",
                        id_values=uuids_to_link,
                        entity=new_entity,
                        shortning=entity_shortning,
                    )
                    metadata = {metadata_key: new_entity[metadata_key] for metadata_key in metadata_keys}
                    metadata["embedding_source"] = "description"
                    metadata["doc_source_type"] = "generated"
                    db_manager.vector_db.add(
                        documents=[new_entity[embedding_key]],
                        metadatas=[metadata],
                        ids=[uuid_],
                    )
                else:
                    print("No similar entities found for entity type: ", entity_type)
            except Exception as e:
                print(f"Error when aligning entities from {uuid}: {e}")
                continue


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    main(args)
